#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/atomic.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "lcd.h"
#include "keys.h"
#include "elmcie.h"
#include "ext_strings.h"
#include "delay_x.h"		// delay lib from Hans-Juergen Heinrichs


extern volatile u8 key_state;		// debounced and inverted key state:
									// bit = 1: key pressed
extern volatile u8 key_press;		// key press detect

extern volatile u8 key_rpt;			// key long press and repeat


// Interrupt routine for timer 0
// Timer 0 is used to debounce both keys.
// Called every 11.1 ms to avoid timer 0
// and timer2 interrupting at the same time.
ISR(TIMER0_COMPA_vect)
{
	Read_keys();
}

// Interrupt routine for Timer2
// Time base for both Elm and Elcie.
// Decrements Cnt_10ms each 10 ms
// When Cnt_10ms reaches 0, it stops TMR1 and TMR2
ISR(TIMER2_COMPA_vect)
{
	Cnt_10ms--;

	if (Cnt_10ms == 0)
	{
		TCCR1B &= ~((1<<CS10)|(1<<CS11)|(1<<CS12));
		TCCR2B &= ~((1<<CS20)|(1<<CS21)|(1<<CS22));
	}

}

//****************************************************************************
// Auxiliary routines

// mimics BASCOM SOUND function
void sound(volatile uint8_t* port, uint8_t mask, uint16_t duration, uint16_t pulses) {

	// Delay function based on AVR-libc "_delay_loop_2"
	// with 2 NOPs included to match BASCOMs delay inside
	// SOUND function
	void my_delay(uint16_t __count) {
		__asm__ volatile (
			"1: sbiw %0,1"	"\n\t"
			"nop"			"\n\t"
			"nop"			"\n\t"
			"brne 1b"
			: "=w" (__count)
			: "0" (__count)
		);}

	while (duration-- > 0) {
		*port |= (mask);
		my_delay(pulses);
		*port &= ~(mask); 
		my_delay(pulses);
	}
}


void Do_error_sound_and_display (void)
{
	sound(&PORTC, _BV(PC4), 115, 11520);
	LcdPrintString_P(ErrorString);
	_delay_s(1);
}

void Do_okay_sound (void)
{
	sound(&PORTC, _BV(PC4), 12, 5760);
	_delay_ms(100);
	sound(&PORTC, _BV(PC4), 35, 2880);
}

void Do_cal_display (void)
{
	LcdClear();
	sound(&PORTC, _BV(PC4), 115, 1152);
	LcdPrintString_P(CalString);
	LcdGotoXY(0,1);
	LcdPrintString_P(PNullString);
	while (! get_key_press(1<<KEY2));		// Waits for null button to be pressed
	_delay_ms(200);
	LcdClear();
	sound(&PORTC, _BV(PC4), 115, 1152);
	if (Elm_selected) {						// Prints the correct string for Elm or Elcie
		LcdPrintString_P(CalElmString);
	}
	else {
		LcdPrintString_P(CalElcieString);
	}
	LcdGotoXY(0,1);
	LcdPrintString_P(PNullString);
	while (! get_key_press(1<<KEY2));		// Waits again for null button
	_delay_ms(200);
	sound(&PORTC, _BV(PC4), 115, 1152);
}

void Do_display(void)
{
	LcdClear();

	// If wrong result, print question marks 
	if (Teh_result == -10000) {
		LcdPrintString_P(WrongResString);
	}
	else
	{
		// Scale the result and converts from float to string
		if ((Unit == pF) || (Unit == uH)) {
			dtostrf(Teh_result, 3, 1, Teh_result_str);
		}
		else {
			if (Teh_result < 10) {
				dtostrf(Teh_result, 3, 3, Teh_result_str);
			}
			else if (Teh_result < 100) {
				dtostrf(Teh_result, 3, 2, Teh_result_str);
			}
			else if (Teh_result < 1000) {
				dtostrf(Teh_result, 3, 1, Teh_result_str);
			}
			else {
				dtostrf(Teh_result, 3, 0, Teh_result_str);
			}
		}

		// Prints the result and it's unit
		LcdPrintString(Teh_result_str);
		LcdPrintString_P(UnitString[Unit]);
	}

	LcdGotoXY(11, 0);
	if (Elm_selected) {
		LcdPrintString("Elm");
		if (!Elm_large_caps) {
			LcdPrintString("Lo");
		}
		else {
			LcdPrintString("Hi");
		}
	}
	else {
		LcdPrintString("LC");
		if (!Elcie_cap) {
			LcdPrintString("ind");
		}
		else {
			LcdPrintString("cap");
		}
	}

	_delay_ms(500);

}
//****************************************************************************



// Get ADC value from current selected channel
int GetADC(void) 
{
	ADCSRA |= (1<<ADSC);

	while bit_is_set(ADCSRA, ADSC);

	return (ADCL | (ADCH<<8));
}

void Teh_term(void)
{
	Tmp_long = Elcie_f_null >> 4;
	Elcie_f_dut = Elcie_f_dut >> 4;
	Tmp_long = Tmp_long * Tmp_long;
	Elcie_f_dut = Elcie_f_dut * Elcie_f_dut;
	Tmp_single = (float)Tmp_long / Elcie_f_dut;
	Tmp_single = Tmp_single - 1;
}

// Handles the Elm part of Elmcie.
// Sets Elm status bits (Ready, Timeout and Missed_it) accordingly
// Returns the counter values, Elm_high and Total_Counts
void Do_Elm(void)
{
	// Discharge DUT
	Elm_low_threshold = 1;

	Status_copy = Status;				// Remember the current setting
	N_elm3k3_lm311on = 1;				// disable the pull-up
	while (bit_is_set(ACSR, ACO))		// by waiting for ACO to go low
	{
		Elm_disch = 1;					// discharge the dut
		_delay_ms(10);
		Elm_disch = 0;					// stop the discharge to check
		_delay_us(5);
	}
	Status = Status_copy;				// Restore the current setting
	Elm_disch = 1;

	// Set the pull-up according to the requested range, and add some more to the discharge time
	// to make sure we are definitely under the low threshold (important for large cap's)
	if (!Elm_large_caps)
	{
		N_elm3k3_lm311on = 1;
		_delay_ms(10);
	}
	else
	{
		N_elm3k3_lm311on = 0;
		_delay_ms(100);
	}

	ACSR = (1<<ACIS0)|(1<<ACIS1);		// Interupt in the positive edge
	ACSR |= (1<<ACI);					// Clear the interrupt flag
	ACSR |= (1<<ACIE);					// Enable comparator interrupt
	_delay_ms(5);

	// use timer 2 as a timeout - 4 s
	Cnt_10ms  = 400;
	TCNT2 = 0;
	TCCR2B |= (1<<CS20)|(1<<CS21)|(1<<CS22);	// Start Timer 2;

	// Reset variables
	Elm_ready = 0;
	Elm_timeout = 0;
	Elm_missed_it = 0;
	Elmcie.Counts = 0;
	Elm_high.Counts = 0;
	TCNT1 = 0;

	// Just copying Nard's code, below. I will comment the following line
	// to check if the modification above, in the ACSR reg, will help.
	ACSR |= (1<<ACI);					// Re-clear a possible flag

	// Start Timer 1 in Elm mode
	TCCR1B |= (1<<CS10);
	// Start the measurement by releasing the discharge
	Elm_disch = 0;

	while (!Elm_ready)
	{
		ATOMIC_BLOCK(ATOMIC_FORCEON)
		{
			if (Cnt_10ms == 0)
			{
				Elm_timeout = 1;
				Elm_ready = 1;
			}
		}
		if bit_is_set(ACSR, ACO)
		{
			Elm_missed_it = 1;
		}
	}
	
	// Stop Timer 1 and 2
	TCCR1B &= ~((1<<CS10)|(1<<CS11)|(1<<CS12));
	TCCR2B &= ~((1<<CS20)|(1<<CS21)|(1<<CS22));
}

void Do_elcie (void)
{
	uint16_t cnt_copy;

	Cnt_10ms = 100;
	TCNT2 = 0;
	Elmcie.Counts = 0;
	TCNT1 = 0;
	GTCCR |= (1<<PSRASY);
	TCCR2B |= (1<<CS20)|(1<<CS21)|(1<<CS22);
	TCCR1B |= (1<<CS10)|(1<<CS11)|(1<<CS12);
	do
	{
		ATOMIC_BLOCK(ATOMIC_FORCEON)
		{
			cnt_copy = Cnt_10ms;
		}
	}
	while (cnt_copy != 0);

	Elmcie.ints[0] = TCNT1;
}


int __attribute__((OS_main)) main (void)
{
	DDRC  = 0x13;
	PORTC = 0x0C;		// SW1 & SW2 pull-ups 
	
	DDRB  = 0x3F;		// All pins output
	PORTB = 0x02;

	DDRD  = 0x5A;
	PORTD = 0x04;		// Pull-up on D2 - Holds power on on original Elmcie

	ADCSRA = 0x83;		// ADC enable - PS = 128 - f = 144 kHz
	ADMUX = 0x45;		// Select ADC5 - Ref = AVcc + ext cap
	GetADC();			// Discard first conversion

	
	// Timer 0 in CTC mode, prescaler = 1024
	// OCR = 0xFF, giving a 14.2222 ms interrupt interval
	// This timer is used to check the switches status
	TCCR0A = (1<<WGM01);
	TCCR0B = (1<<CS02)|(1<<CS00);
	OCR0A  = 0xC7;
	TIMSK0 = (1<<OCIE0A);

	// Timer 1 starts counting pulses from LM311 (Elcie) when
	// TCCR1B |= (1<<CS10)|(1<<CS11)|(1<<CS12)
	// For Elm (counts pulses from main clock - 18.432 MHz):
	// TCCR1B |= (1<<CS10)

	// Set Timer 1 Overflow interrupt
	TIMSK1 = (1<<TOIE1);
	
	// Timer 2 in CTC mode, prescaler = 1024
	// OCR = 0xB3 = 10 ms
	// Start timer -> TCCR2B |= (1<<CS20)|(1<<CS21)|(1<<CS22)
	// Stop timer -> TCCR2B &= ~((1<<CS20)|(1<<CS21)|(1<<CS22))
	TCCR2A = (1<<WGM21)|(0<<WGM20);
	TCCR2B = (1<<WGM22);
	OCR2A  = 0xB3;

	// Set Timer 2 OC2A interrupt
	TIMSK2 = (1<<OCIE2A);

	LcdInitialize();

	sei();

	// Restore from eeprom
	Elm_high_range_cal = eeprom_read_dword((void*)&Ee_elm_high_range_cal);
	Elm_low_range_cal = eeprom_read_dword((void*)&Ee_elm_low_range_cal);
	C_elcie_null = eeprom_read_dword((void*)&Ee_c_elcie_null);
	L_elcie_null = eeprom_read_dword((void*)&Ee_l_elcie_null);
	Elcie_f_null = eeprom_read_dword(&Ee_elcie_f_null);
	Elm_null_cnts = eeprom_read_dword(&Ee_elm_null_cnts);

	LcdPrintString_P(ElmcieString);
	LcdGotoXY(0,1);

	// Battery/power check
	// Under full load, the minimun allowed voltage is 7.35V = 270 counts
	// Less means that the 7805 will drop out of regulation
	Tmp_word = GetADC();

	if (Tmp_word < 270)
	{
		LcdPrintString_P(BatLowString);
	}
	else
	{
		// max current
		N_elm3k3_lm311on = 0;
		Backlight_hi = 1;
		Do_l = 1;
		Tmp_word = GetADC();					// Check under max. current draw
		
		Tmp_word = ((Tmp_word - 250) >> 4);
		LcdPrintString_P(BatQString);
		itoa(Tmp_word, Teh_result_str, 10);
		LcdPrintString(Teh_result_str);
	}

	// Set for min. current consumption
	N_elm3k3_lm311on = 1;
	Backlight_hi = 0;
	Do_l = 0;

	_delay_s(2);

	// Default selected
	Elm_selected = 1;

	// Default select for Elcie
	Elcie_cap = 1;

	LcdClear();

	for(;;)
	{
		
		// ********************************************************************************
		// Check the keys and sets the status bits according to the required function
		// short SW1 - toggle Elcie C/Elcie L
		// long  SW1 - toggle Elcie / Elm
		// short SW2 - Null
		// Long  SW2 - Calibration

		if (get_key_short(1<<KEY2))
		{
			Do_nulling = 1;
		}
		if (get_key_long(1<<KEY2))
		{
			Do_calibration = 1;
		}

		if (get_key_short(1<<KEY1))
		{
			if (Elm_selected == 0){
				Elcie_cap ^= 1;		// toggle between Elcie-capacitive and inductive
			}
		}
		if (get_key_long(1<<KEY1))
		{
			Elm_selected ^= 1;		// toggle between Elm and Elcie
		}

		if ((Elm_selected == 0) && (Elcie_cap == 0))
		{
			Do_l = 1;
		}
		else {
			Do_l = 0;
		}



//*************************************************************************
// Elm Mode

		if (Elm_selected)
		{
			// User wants zero pf on the display
			if (Do_nulling)
			{
				LcdClear();
				sound(&PORTC, _BV(PC4), 115, 1152);
				LcdPrintString_P(RemDutString);
				LcdGotoXY(0,1);
				LcdPrintString_P(PNullString);

				while (! get_key_press(1<<KEY2));

				sound(&PORTC, _BV(PC4), 115, 1152);
				_delay_ms(200);
				LcdClear();
				LcdPrintString_P(NullingString);

				Elm_null_cnts = 0;

				LcdGotoXY(0, 1);

				// Read the counts 8 times
				for (Tmp_byte = 8; (Tmp_byte > 0); Tmp_byte--)
				{
					Do_Elm();
					Elm_cnt = Elmcie.Counts - Elm_high.Counts;
					Elm_null_cnts = Elm_null_cnts + Elm_cnt;
					LcdSendCharacter('.');
					_delay_ms(200);
				}
				Elm_null_cnts = Elm_null_cnts >> 3; // and average dividing by 8

				// if its outside of good range is an error.
				if ((Elm_null_cnts > Elm_null_cnts_max)	|| \
					(Elm_null_cnts < Elm_null_cnts_min))		
				{
					Do_error_sound_and_display();		
					Elm_null_cnts = eeprom_read_dword(&Ee_elm_null_cnts);	// get the last know value from EEprom
				}
				else
				{
					eeprom_write_dword(&Ee_elm_null_cnts, Elm_null_cnts);	// if its good, store it
					LcdPrintString_P(DoneString);
					Do_okay_sound();
				}

				Do_nulling = 0;

				_delay_s(1);
				LcdClear();
			}
		
			if (Do_calibration)
			{
				Do_cal_display();
				
				LcdClear();
				Tmp_long = 0;
				for (Tmp_byte = 8; (Tmp_byte > 0); Tmp_byte--)
				{
					Do_Elm();
					Elm_cnt = Elmcie.Counts - Elm_high.Counts;
					Elm_cnt = Elm_cnt - Elm_null_cnts;
					Tmp_long = Tmp_long + Elm_cnt;
					LcdSendCharacter('.');
					_delay_ms(200);
				}
				Tmp_long = Tmp_long >> 3;
				Elm_low_range_cal = (float)Tmp_long / C_cal1;

				if ((Elm_low_range_cal < Elm_low_range_cal_min) || \
					(Elm_low_range_cal > Elm_low_range_cal_max))
				{
					Do_error_sound_and_display();
					Elm_low_range_cal = eeprom_read_dword((void*)&Ee_elm_low_range_cal);
				}
				else
				{
					eeprom_write_dword((void*)&Ee_elm_low_range_cal,(uint32_t)Elm_low_range_cal);
					Do_okay_sound();
					LcdPrintString_P(ElmLowString);
				}
				LcdGotoXY(0,1);
				Elm_large_caps = 1;
				Tmp_long = 0;
				for (Tmp_byte = 8; (Tmp_byte > 0); Tmp_byte--)
				{
					Do_Elm();
					Elm_cnt = Elmcie.Counts - Elm_high.Counts;
					Tmp_long = Tmp_long + Elm_cnt;
					LcdSendCharacter('.');
					_delay_ms(200);
				}
				Tmp_long = Tmp_long >> 3;
				Elm_high_range_cal = (float)Tmp_long / C_cal1;

				if ((Elm_high_range_cal < Elm_high_range_cal_min) || \
					(Elm_high_range_cal > Elm_high_range_cal_max))
				{
					Do_error_sound_and_display();
					Elm_high_range_cal = eeprom_read_dword((void*)&Ee_elm_high_range_cal);
				}
				else
				{
					eeprom_write_dword((void*)&Ee_elm_high_range_cal,(uint32_t)Elm_high_range_cal);
					Do_okay_sound();
					LcdPrintString_P(ElmHighString);
				}

				Elm_large_caps = 0;
				Do_calibration = 0;
				_delay_s(1);
				LcdClear();
			}

			Do_Elm();

			if (!Elm_timeout)
			{
				
				// Calculate the number or clocks between low and high tresholds
				Elm_cnt = Elmcie.Counts - Elm_high.Counts;
				
				if (!Elm_large_caps)
				{
					// Subtract null counts in low range
					if (Elm_cnt > Elm_null_cnts)
					{
						Elm_cnt = Elm_cnt - Elm_null_cnts;
					}
					else
					{
						Elm_cnt = 0;
					}

					Teh_result = (float)Elm_cnt / Elm_low_range_cal;
				}
				else
				{
	            	// no null-counts in high range
            		Teh_result = (float)Elm_cnt / Elm_high_range_cal;		//from counts to picoFarads
            		
					// if the counts are very low when in high range mode, switch back to low range mode
	            	if (Elm_cnt < 30000)
					{
    	           		Elm_large_caps = 0;
					}
				}

				if (Teh_result < 1000)
				{
					Unit = pF;
				}
				else if (Teh_result < 400000)
				{
					Teh_result /= 1000;
					Unit = nF;
				}
				else
				{
					Teh_result /= 1000000;
					Unit = uF;
				}
			}
			else
			{
				// Timeout is set. Enable high range mode
				if (!Elm_large_caps)
				{
					Elm_large_caps = 1;
				}
				else
				{
					// timeout in high range mode. Out of range
					Do_error_sound_and_display();
					Teh_result = -10000;
					Unit = 10;
					Elm_large_caps = 0;
					_delay_ms(200);
				}
			}

		}

//*************************************************************************
// Elcie Mode

		else
		{
			N_elm3k3_lm311on = 0;

			if (Do_nulling)
			{
				LcdClear();
				sound(&PORTC, _BV(PC4), 115, 1152);

				if (Elcie_cap)
				{
					LcdPrintString_P(RemDutString);
				}
				else
				{
					LcdPrintString_P(BridgeDutString);
				}

				LcdGotoXY(0,1);
				LcdPrintString_P(PNullString);

				while (! get_key_press(1<<KEY2));

				sound(&PORTC, _BV(PC4), 115, 1152);
				_delay_ms(200);
				LcdClear();
				LcdPrintString_P(NullingString);
				LcdGotoXY(0,1);
				
				Elcie_f_null = 0;
				for (Tmp_byte = 4; (Tmp_byte > 0); Tmp_byte--)
				{
					Do_elcie();
					Elcie_f_null += Elmcie.Counts;
					LcdPrintString("..");
					_delay_ms(100);
				}
				Elcie_f_null /= 4;
				Do_nulling = 0;

				if ((Elcie_f_null > Elcie_f_null_max) || (Elcie_f_null < Elcie_f_null_min))
				{
					Do_error_sound_and_display();
					Elcie_f_null = eeprom_read_dword((void*)&Ee_elcie_f_null);
				}
				else
				{
					eeprom_write_dword((void*)&Ee_elcie_f_null,(uint32_t)Elcie_f_null);
					Do_okay_sound();
				}

				_delay_s(1);
				LcdClear();
			}

			if (Do_calibration)
			{
				Do_cal_display();

				LcdClear();
				Elcie_f_dut = 0;
				LcdPrintString("Calibrating .. ");
				LcdGotoXY(0,1);
				for (Tmp_byte = 4; (Tmp_byte > 0); Tmp_byte--)
				{
					Do_elcie();
					Elcie_f_dut += Elmcie.Counts;
					LcdPrintString("..");
					_delay_ms(100);
				}
				Elcie_f_dut /= 4;

				Teh_term();
				Tmp_long = 1000000000UL / Elcie_f_null;
				Tmp_long = Tmp_long * Tmp_long;
				L_elcie_null = Tmp_single * Tmp_long;
				L_elcie_null = L_elcie_null * Factor2;
				L_elcie_null = L_elcie_null / C_cal2;
				L_elcie_null = L_elcie_null / 10000;

				if ((L_elcie_null > L_elcie_null_max) || (L_elcie_null < L_elcie_null_min))
				{
					Do_error_sound_and_display();
					L_elcie_null = eeprom_read_dword((void*)&Ee_l_elcie_null);
				}
				else
				{
					eeprom_write_dword((void*)&Ee_l_elcie_null,(uint32_t)L_elcie_null);
				}

				C_elcie_null = C_cal2 / Tmp_single;
				C_elcie_null = C_elcie_null / 10;

				if ((C_elcie_null > C_elcie_null_max) || (C_elcie_null < C_elcie_null_min))
				{
					Do_error_sound_and_display();
					C_elcie_null = eeprom_read_dword((void*)&Ee_c_elcie_null);
				}
				else
				{
					eeprom_write_dword((void*)&Ee_c_elcie_null,(uint32_t)C_elcie_null);
					Do_okay_sound();
				}

				Do_calibration = 0;

				_delay_s(1);
				LcdClear();
			}

			Do_elcie();

			if (Elcie_cap)
			{
				Elcie_f_dut = Elmcie.Counts;
				Teh_term();
				C_elcie_dut = Tmp_single * C_elcie_null;

				if (C_elcie_dut < 1000)
				{
					Teh_result = C_elcie_dut;
					Unit = pF;
				}
				else
				{
					Teh_result = C_elcie_dut / 1000;
					Unit = nF;
				}
			}
			else
			{
				Elcie_f_dut = Elmcie.Counts;
				Teh_term();
				L_elcie_dut = Tmp_single * L_elcie_null;

				if (Elmcie.Counts < 10)
				{
					Teh_result = -10000;
				}
				else
				{
					if (L_elcie_dut < 1000 )
					{
						Teh_result = L_elcie_dut;
						Unit = uH;
					}
					else
					{
						Teh_result = L_elcie_dut / 1000;
						Unit = mH;
					}
				}
			}

		}

		Do_display();

	}
	
}
