//*****************************************************************************
//
// File Name	: 'keys.h'
// Title		: Debounce and repeat keys
// Author		: Peter Dannegger (code) Felipe Maimon (lib)
// Created		: 2009-03-06
// Target MCU	: Atmel AVR series
//
//
// Peter Dannegger's code to debounce and repeat keys. Up to 8 keys
// (at the same port) can be debounced at once. Just create an interrupt
// triggered every 10 ms and call the Read_Keys function from it.
//
//*****************************************************************************

#ifndef KEYS_H
#define KEYS_H

typedef unsigned char	u8;
typedef signed short	s16;


#define KEY_PIN		PINC
#define KEY1		3
#define KEY2		2


#define REPEAT_MASK		(1<<KEY1|1<<KEY2)	// repeat: key1, key2
#define REPEAT_START	50					// after 500ms
#define REPEAT_NEXT		20					// every 200ms


volatile u8 key_state;				// debounced and inverted key state:
									// bit = 1: key pressed
volatile u8 key_press;				// key press detect

volatile u8 key_rpt;				// key long press and repeat

static inline void Read_keys (void) __attribute__((always_inline));

u8 get_key_press( u8 key_mask );

u8 get_key_rpt( u8 key_mask );

u8 get_key_short( u8 key_mask );

u8 get_key_long( u8 key_mask );


// **************************************************************************
// This function needs to be in the header file so it can be inlined
void Read_keys (void) {

	static u8 ct0, ct1, rpt;
	u8 i;

	i = key_state ^ ~KEY_PIN;		// key changed ?
	ct0 = ~( ct0 & i );				// reset or count ct0
	ct1 = ct0 ^ (ct1 & i);			// reset or count ct1
	i &= ct0 & ct1;					// count until roll over ?
	key_state ^= i;					// then toggle debounced state
	key_press |= key_state & i;		// 0->1: key press detect

	if( (key_state & REPEAT_MASK) == 0 )	// check repeat function
		rpt = REPEAT_START;			// start delay
	if( --rpt == 0 ){
		rpt = REPEAT_NEXT;			// repeat delay
		key_rpt |= key_state & REPEAT_MASK;
	}
}


#endif
