/*! \file lcd.c \brief Character LCD driver for HD44780/SED1278 displays. */
//*****************************************************************************
//
// File Name	: 'lcd.C'
// Title		: Character LCD driver for HD44780/SED1278 displays
// Author		: Felipe Maimon
// Created		: 06/18/2008
// Target MCU	: Atmel AVR series
//
// This code is distributed under the GNU Public License
//		which can be found at http://www.gnu.org/licenses/gpl.txt
//
//*****************************************************************************

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "lcd.h"


// Macros to help setting up the necessary nibbles to 4 bit
// communication
#define Clear_Nibble(addr, shift)		addr &= ~(0x0F<<shift)
#define Set_Nibble(addr, shift)			addr |= (0x0F<<shift)


/*************************************************************/
/********************** LOCAL FUNCTIONS **********************/
/*************************************************************/

/* LcdSendNibble
 * 
 * Sends a 4-bit nibble to the display.
 *
 * Parameters:
 *	uint8_t nibble	The high nibble of of this byte
 *					is sent to the display.
 * Returns:
 *	nothing
 */
void LcdSendNibble( uint8_t nibble )
{
   _delay_ms(1.7);

   // Output upper nibble on the data ports upper bits
   Clear_Nibble(LCD_DATA_POUT, LCD_DATA_START);
   LCD_DATA_POUT |= (nibble&0xF0)>>(4-LCD_DATA_START);

   // Toggle the E line
   LCD_CTRL_PORT |= (1<<LCD_E);   // Going up..
   _delay_us(1);
   LCD_CTRL_PORT &= ~(1<<LCD_E);  // ..and down.
}

/* LcdSendByte
 * 
 * Sends a 8-bit byte to the display.
 *
 * Parameters:
 *	uint8_t theByte	The byte to send to the display
 *
 * Returns:
 *	nothing
 */
void LcdSendByte(uint8_t theByte)
{
   // Send the high nibble
   LcdSendNibble(theByte);

   // Shift theByte to get lower nibble in upper part...
   theByte = theByte << 4;
   // ...and send it
   LcdSendNibble(theByte);

}

/* LcdSendInstruction
 * 
 * Sends an instruction to the display.
 *
 * Parameters:
 *	uint8_t command	This byte is sent to the display as 
 *					an instruction (RS low).
 * Returns:
 *	nothing
 */
void LcdSendInstruction( uint8_t theInstruction )
{
   // RS low for instructions
   LCD_CTRL_PORT &= ~(1<<LCD_RS);

   // Send the instruction
   LcdSendByte(theInstruction);
}

/* LcdSendCharacter
 * 
 * Sends a character to the display.
 *
 * Parameters:
 *	uint8_t nibble	This byte is sent to the display as 
 *					a character (RS high).
 * Returns:
 *	nothing
 */
void LcdSendCharacter(uint8_t theChar)
{
   // RS high for characters to display
   LCD_CTRL_PORT |= (1<<LCD_RS);

   // Send the command
   LcdSendByte(theChar);
}



/*************************************************************/
/********************* PUBLIC FUNCTIONS **********************/
/*************************************************************/

void LcdInitialize()
{
   // initialize LCD control lines
   LCD_CTRL_PORT &= ~(1<<LCD_RS);	// RS low
   LCD_CTRL_PORT &= ~(1<<LCD_E);	// E low

   // initialize LCD control lines to output
   LCD_CTRL_DDR |= (1<<LCD_RS);
   LCD_CTRL_DDR |= (1<<LCD_E);

   // initialize LCD data port to input
   LCD_DATA_DDR |= 0xF0>>LCD_DATA_START;	// Data on high four bits of port for now...

   // First part of init sequence is 3 x Function Set with
   // stipulated waits. Note that the display is in 8-bit mode
   // initially, but although the four low data lines are not connected
   // this does not matter as the instructions low nibble is zero anyway.
   _delay_ms(15);
   LcdSendNibble( LCD_FUNCTION_SET );
   _delay_ms(5);
   LcdSendNibble( LCD_FUNCTION_SET );
   _delay_us(100);
   LcdSendNibble( LCD_FUNCTION_SET );

   // Now, still in 8-bit mode, set the display to 4-bit mode
   LcdSendNibble( LCD_FUNCTION_SET_4BIT );

   // We are now in 4-bit mode.
   // Do the rest of the init sequence.
   LcdSendInstruction( LCD_FUNCTION_SET_4BIT );
   LcdSendInstruction( LCD_DISPLAY_OFF );
   LcdSendInstruction( LCD_DISPLAY_CLEAR );
   LcdSendInstruction( LCD_ENTRY_MODE_SET );
   LcdSendInstruction( LCD_DISPLAY_ON );
}
/*
void LcdHome(void)
{
	// move cursor to home
	LcdSendInstruction(LCD_CURSOR_HOME);
}
*/
void LcdClear(void)
{
	// clear LCD
	LcdSendInstruction(LCD_DISPLAY_CLEAR);
}

void LcdGotoXY(uint8_t x, uint8_t y)
{
	uint8_t DDRAMAddr;

	// remap lines into proper order
	switch(y)
	{
	case 0: DDRAMAddr = LCD_LINE0_DDRAMADDR+x; break;
	case 1: DDRAMAddr = LCD_LINE1_DDRAMADDR+x; break;
	case 2: DDRAMAddr = LCD_LINE2_DDRAMADDR+x; break;
	case 3: DDRAMAddr = LCD_LINE3_DDRAMADDR+x; break;
	default: DDRAMAddr = LCD_LINE0_DDRAMADDR+x;
	}

	// set data address
	LcdSendInstruction(LCD_DDRAM | DDRAMAddr);
}


/*void LcdPrintData(char* data, uint8_t nBytes)
{
	register uint8_t i;

	// check to make sure we have a good pointer
	if (!data) return;

	// print data
	for(i=0; i<nBytes; i++)
	{
		LcdSendCharacter(data[i]);
	}
}
*/
void LcdPrintString(char* data)
{
	register uint8_t i = 0;

	// check to make sure we have a good pointer
	if (!data) return;

	// print data
	while (data[i] != 0)
	{
		LcdSendCharacter(data[i]);
		i++;
	}
}

void LcdPrintString_P(const char *data)
{
	// print data
	while (pgm_read_byte(data) != 0x00)
	{
		LcdSendCharacter(pgm_read_byte(data++));
	}
}

