#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <stdlib.h>

#include <util/delay.h>

#include "lcd.h"
#include "elmcie.h"


// Interrupt routine for timer 0
// Timer 0 is used to check how long the swithces were
// pressed. Called every 14.2 ms to avoid timer 0 and
// timer2 interrupting at the same time.
// Switches are debounced by checking if it were pressed
// by at least three consecutive times
//
// ATTENTION! User have to clear the switch status flags after reading it!!!
//
ISR(TIMER0_COMPA_vect)
{
	if (SW1_n) {
		if (Switches.SW1_cur_value > 2) {
			Switches.SW1_released = 1;
			Switches.SW1_last_value = Switches.SW1_cur_value;
		}
		Switches.SW1_cur_value = 0;
	}
	else {
		if (Switches.SW1_cur_value < 255)
			Switches.SW1_cur_value++;
	}

	if (SW2_n) {
		if (Switches.SW2_cur_value > 2) {
			Switches.SW2_released = 1;
			Switches.SW2_last_value = Switches.SW2_cur_value;
		}
		Switches.SW2_cur_value = 0;
	}
	else {
		if (Switches.SW2_cur_value < 255)
			Switches.SW2_cur_value++;
	}
}

// Interrupt routine for timer1.
// Just increases Elmcie_Count to achieve higher
// resolution, 16 from the timer and 16 from the
// variable, giving a 32 bits value.
ISR(TIMER1_OVF_vect)
{
	Elmcie_Count++;
}

// Interrupt routine for Timer2
// Time base for both Elm and Elcie.
// Decrements Cnt_10ms each 10 ms
// When Cnt_10ms reaches 0, it stops TMR1 and TMR2
ISR(TIMER2_COMPA_vect)
{
	Cnt_10ms--;

	if (Cnt_10ms == 0) {
		TCCR1B &= ~((1<<CS10)|(1<<CS11)|(1<<CS12));
		TCCR2B &= ~((1<<CS20)|(1<<CS21)|(1<<CS22));
	}

}


// Interrupt routine for the analog comparator used in Elm
// May need to rewrite in asm.
// Generated code with lots of pushes and pops
ISR(ANALOG_COMP_vect)
{
	if bit_is_set(DDRD,PD4) {
		Elm_low_threshold = 0;
		Elm_high = Elmcie_Count;
		Elm_high = TCNT1 | (Elm_high << 16);
		ACSR |= (1<<ACI);
	}
	else {
		Total_Counts = TCNT1;
		Elm_low_threshold = 1;
		ACSR &= ~(1<<ACIE);
		ACSR |= (1<<ACI);
	}
}

// Get ADC value from current selected channel
int GetADC(void) 
{
	ADCSRA |= (1<<ADSC);

	while bit_is_set(ADCSRA, ADSC);

	return (ADCL | (ADCH<<8));
}

// Initialize all ports, timers, adc, etc...
void initHW (void)
{

	DDRC  = 0x13;
	PORTC = 0x0C;		// SW1 & SW2 pull-ups 
	
	DDRB  = 0x3F;		// All pins output
	PORTB = 0x02;

	DDRD  = 0x5A;
	PORTD = 0x04;		// Pull-up on D2 - Holds power on on original Elmcie

	ADCSRA = 0x83;		// ADC enable - PS = 128 - f = 144 kHz
	ADMUX = 0x45;		// Select ADC5 - Ref = AVcc + ext cap
	GetADC();			// Discard first conversion

	
	// Timer 0 in CTC mode, prescaler = 1024
	// OCR = 0xFF, giving a 14.2222 ms interrupt interval
	// This timer is used to check the switches status
	TCCR0A = (1<<WGM01);
	TCCR0B = (1<<CS02)|(1<<CS00);
	OCR0A  = 0xFF;
	TIMSK0 = (1<<OCIE0A);

	// Timer 1 starts counting pulses from LM311 (Elcie) when
	// TCCR1B = (1<<CS10)|(1<<CS11)|(1<<CS12)
	// For Elm (counts pulses from main clock - 18.432 MHz):
	// TCCR1B = (1<<CS10)

	// Set Timer 1 Overflow interrupt
	TIMSK1 = (1<<TOIE1);
	
	// Timer 2 in CTC mode, prescaler = 1024
	// OCR = 0xB3 = 10 ms
	// Start timer -> TCCR2B |= (1<<CS20)|(1<<CS21)|(1<<CS22)
	// Stop timer -> TCCR2B &= ~((1<<CS20)|(1<<CS21)|(1<<CS22))
	TCCR2A = (1<<WGM21)|(0<<WGM20);
	TCCR2B = (1<<WGM22);
	OCR2A  = 0xB3;

	// Set Timer 2 OC2A interrupt
	TIMSK2 = (1<<OCIE2A);

	LcdInitialize();

}


int main (void)
{
	char Out_Str[8];

	initHW();

	sei();

	while (1) {

/*		Cnt_10ms = 100;
	
		Elmcie_Count = 0;
		TCNT1 = 0;					// Clear Timer 1
		TCNT2 = 0;					// Clear Timer 2
		GTCCR |= (1<<PSRASY);		// Clear pre-scaler

		TCCR2B |= (1<<CS20)|(1<<CS21)|(1<<CS22);	// Start Timer2
		TCCR1B = (1<<CS10)|(1<<CS11)|(1<<CS12);		// Count external pulses

		while (Cnt_10ms) ;				// Wait 1 second

		Total_Counts = Elmcie_Count;
		Total_Counts = TCNT1 | (Total_Counts << 16);

*/

		LcdHome();
		if (!SW1_n) {
			LcdPrintData("SW1 On  ", 8);
		}
		else {
			LcdPrintData("SW1 Off ", 8);
		}
		if (Switches.SW1_released) {
			Switches.SW1_released = 0;
			itoa(Switches.SW1_last_value, Out_Str, 10);
			LcdPrintData(Out_Str, 3);
		}

		LcdGotoXY(0,1);
		if (!SW2_n) {
			LcdPrintData("SW2 On  ", 8);
		}
		else {
			LcdPrintData("SW2 Off ", 8);
		}
		if (Switches.SW2_released) {
			Switches.SW2_released = 0;
			itoa(Switches.SW2_last_value, Out_Str, 10);
			LcdPrintData(Out_Str, 3);
		}
	}
	
	return 0;
}
