/*! \file lcd.h \brief Character LCD driver for HD44780/SED1278 displays. */
//*****************************************************************************
//
// File Name	: 'lcd.h'
// Title		: Character LCD driver for HD44780/SED1278 displays
// Author		: Felipe Maimon
// Created		: 06/18/2008
// Target MCU	: Atmel AVR series
//
//
// Code based on Pascal Stang's and Johan Ekdahl's LCD driver, modified to fit
//		the needs for elmcie hardware Rev 5.1.
// Supports 4-bit LCD mode, write only. Everything else removed
// Now uses standard C constructs instead of deprecated functions
//		like sbi, cbi, outb and inb
//
// This code is distributed under the GNU Public License
//		which can be found at http://www.gnu.org/licenses/gpl.txt
//
//*****************************************************************************

#ifndef LCD_H
#define LCD_H

#define LCD_CTRL_PORT	PORTC
#define LCD_CTRL_DDR	DDRC
#define LCD_RS			PC1
#define LCD_E			PC0
#define LCD_DATA_POUT	PORTB
#define LCD_DATA_PIN	PINB
#define LCD_DATA_DDR	DDRB
#define LCD_DATA_START	PB2


#define LCD_FUNCTION_SET		0x38	// 0b00110000
#define LCD_FUNCTION_SET_4BIT	0x28	// 0b00101000
#define LCD_DISPLAY_OFF			0x08	// 0b00001000
//#define LCD_DISPLAY_ON	        0x0F	// 0b00001111
#define LCD_DISPLAY_ON		  	0x0C	// 0b00001100
#define LCD_DISPLAY_CLEAR		0x01	// 0b00000001
#define LCD_ENTRY_MODE_SET		0x06	// 0b00000110
#define LCD_CURSOR_HOME			0x02	// 0b00000010
#define LCD_DDRAM				0x80	// 0b10000000

// cursor position to DDRAM mapping
#define LCD_LINE0_DDRAMADDR		0x00
#define LCD_LINE1_DDRAMADDR		0x40
#define LCD_LINE2_DDRAMADDR		0x14
#define LCD_LINE3_DDRAMADDR		0x54

// ****** Low-level functions ******
// the following functions are the only ones which deal with the CPU
// memory or port pins directly.  If you decide to use a fundamentally
// different hardware interface to your LCD, only these functions need
// to be changed, after which all the high-level functions will
// work again.

void LcdSendNibble( uint8_t nibble );
void LcdSendByte(uint8_t theByte);
void LcdSendInstruction( uint8_t theInstruction );
void LcdSendCharacter(uint8_t theChar);


// ****** High-levlel functions ******
// these functions provide the high-level control of the LCD
// such as clearing the display, setting cursor positions,
// displaying text and special characters

// initializes the LCD display (gets it ready for use)
void LcdInitialize(void);

// moves the cursor/position to Home (upper left corner)
void LcdHome(void);

// clears the LCD display
void LcdClear(void);

// moves the cursor/position to the row,col requested
// ** this may not be accurate for all displays
void LcdGotoXY(uint8_t row, uint8_t col);

// prints a series of bytes/characters to the display
void LcdPrintData(char* data, uint8_t nBytes);

#endif
